#include <winsock2.h>
#include <windows.h>
#include <assert.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>

#define SERVER_PORT 8001
#define MAXBAD 8

static BOOL InitWinSock(void)
{
	WORD	wVersionRequested;
	WSADATA	wsaData;
	int		err; 

	/*
	 *	WinSock requires us to call WSAStartup
	 */
	wVersionRequested = MAKEWORD( 2, 0 ); 
	err = WSAStartup( wVersionRequested, &wsaData );
	if ( err != 0 ) {
		/* Tell the user that we couldn't find a usable */
		/* WinSock DLL.                                  */    
		return FALSE;
	}

	return TRUE;
}

typedef enum { SND, RCV } OP;

typedef struct {
  OP op;
  int size;
  int repeat;
  char first;
} CMD;

CMD *cmd;
int SEQCNT;

/*
 * build array for sequence of send/recv commands to issue 
 */
void Build_Seq(char *fname) 
{
  FILE *cmdfile;
  char  buf[100];
  int   r, i, size, repeat;
  char  first;

  cmdfile = fopen(fname,"r");
  if (!cmdfile) {
    fprintf(stderr,"Error. log file <%s> not found\n",fname);
	  exit(-1);
  }
  
  r = fscanf(cmdfile, "%d", &SEQCNT);
  assert(r==1);
  assert(SEQCNT>0 && SEQCNT<1000);
  printf("using command file: %s\nSEQCNT=%d\n",fname,SEQCNT);

  cmd = (CMD*) calloc (1, sizeof(CMD)*SEQCNT);
  assert(cmd);

  for (i=0; i<SEQCNT; i++) {
	  r = fscanf(cmdfile, "%s %d %d %c\n", buf, &size , &repeat, &first);
	  assert(r==4);
	  cmd[i].op     = (stricmp(buf,"send")==0) ? SND : RCV;
	  cmd[i].size   = size;
	  cmd[i].repeat = repeat;
	  cmd[i].first  = first;
  }
}

/*
 * runs the sequence of send/receive commands
 */
char Sbuf[70000], Rbuf[70000];

void Run_Test(int sd) 
{
  int         size, repeat, first;
  int         i, j, k, t, cnt;
  int         errcnt=0;

  printf("starting tests...\n");
  for (i=0; i<SEQCNT; i++) {
	  size   = cmd[i].size;
	  repeat = cmd[i].repeat;
	  first  = cmd[i].first;
	  
	  printf ("Test%d: %s size=%-5d repeat=%-5d first=%c\n",
		  i, (cmd[i].op==SND) ? "send" : "recv",
		  size, repeat, first);
	  fflush(stdout);
	  
	  switch(cmd[i].op) {

	  /* send command */
	  case SND:
	    for (t=0; t<size; t++)
		    Sbuf[t] = first;
	    for (j=0; j<repeat; j++) {
		    cnt = send(sd, Sbuf, size, 0);
		    assert(cnt==size);
	    }
	    break;

	  /* recv command */
	  case RCV:
	    for (t=0; t<size; t++)
		    Rbuf[t] = '*';

	    for (j=0; j<repeat; j++) {
		    cnt = 0;
		    do {
		      cnt += recv(sd, Rbuf+cnt, size-cnt, 0);
        } while (cnt<size && cnt>0);
		    if (cnt < 0) {
		      perror("client calling recv()");
		      exit(-1);
		    } 
		    else if (cnt == 0) {
		      fprintf(stderr, "Connection closed by server\n");
		      exit(-1);
		    } else {
		      for (k=0; k<size; k++) {
			      if (Rbuf[k] != (char)first) {
			        fprintf(stderr,
				        "Got:0x%02x[%1c]  expected:0x%02x[%1c] --> Bad Data at byte %d. size=%d iter=%d\n",
				        0xff&(char)Rbuf[k], 0xff&(char)Rbuf[k], 
				        0xff&(char)first, 0xff&(char)first,
				        k, size, j);
			        if (++errcnt>MAXBAD) {
				        fprintf(stderr,"Program Stopped. Too Many Errors!\n");
				        assert(0);
			        } 
			      }
			      Rbuf[k] = '*';  /* clear */
		      }
		    }
	    }
	    break;
	  } /* switch */
  } /* for all tests */

  if (errcnt)
	  fprintf(stderr,"Oh no, looks BAD, %d errors occured.\n",errcnt);
  else
	  fprintf(stderr,"LOOKS GOOD\n");
}


void main(int argc, char **argv)
{
  struct sockaddr_in serverAddress;
  struct sockaddr_in addr;
  struct hostent *hp;

  int addrLength, sd;
  int isServer;
  int mySocket;

  if (argc < 3) {
    printf("To start server: %s -s cmd_file\n", argv[0]);
    printf("To start client: %s server cmd_file\n", argv[0]);
    exit(0);
  }

  if (strcmp(argv[1], "-s")==0) { /* server */
    isServer=1;
  }
  else { /* client */
    isServer=0;
  }

  Build_Seq(argv[2]);

  /* winsock initialization */
  printf("FD_SETSIZE=%d\n",FD_SETSIZE);
  InitWinSock();

  /* create a socket */
  fprintf(stdout,"calling socket()\n");
	fflush(stdout);
  mySocket = socket(AF_INET, SOCK_STREAM, 0);
  fprintf(stderr,"returned from socket() == %d \n",mySocket);
  if (mySocket == -1) {
	  perror("server calling sn_socket");
	  exit(-1);
  }

  if (isServer) {
    /* bind the socket for server */
    serverAddress.sin_family = AF_INET;
    serverAddress.sin_addr.s_addr = INADDR_ANY;
    serverAddress.sin_port = htons(SERVER_PORT);
    fprintf(stderr,"calling bind()\n");
    if (bind(mySocket, (struct sockaddr*) &serverAddress, sizeof(struct sockaddr_in))) {
	    perror("server calling sn_bind()");
	    exit(-1);
    }

    /* listen for the client */
    listen(mySocket, 5);
  
    addrLength = sizeof(addr);
    sd = accept(mySocket, (struct sockaddr*)&addr, &addrLength);
    if (sd <= 0 ) {
	    perror("server calling sn_accept()");
	    exit(-1);
    }
  }
  else { /* client */
    /* connect to server */
    sd=mySocket;
    hp = gethostbyname(argv[1]);
  	serverAddress.sin_family = AF_INET;
    serverAddress.sin_port = htons(SERVER_PORT);
    memcpy((char *) &serverAddress.sin_addr, hp->h_addr, sizeof(serverAddress.sin_addr));
    if (connect(sd, (struct sockaddr*)&serverAddress, sizeof(struct sockaddr_in))) {
		fprintf(stderr, "client calling sn_connect returned %d :: %s\n",
			WSAGetLastError(), strerror(WSAGetLastError()));
  	  exit(-1);
    }
  }

  Run_Test(sd);

  getchar();
  close(sd);
}
