#include <winsock2.h>
#include <windows.h>
#include <assert.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>

#define SERVER_PORT 8001

static BOOL InitWinSock(void)
{
	WORD	wVersionRequested;
	WSADATA	wsaData;
	int		err; 

	/*
	 *	WinSock requires us to call WSAStartup
	 */
	wVersionRequested = MAKEWORD( 2, 0 ); 
	err = WSAStartup( wVersionRequested, &wsaData );
	if ( err != 0 ) {
		/* Tell the user that we couldn't find a usable */
		/* WinSock DLL.                                  */    
		return FALSE;
	}

	return TRUE;
}

#define MAXBUFSIZE 1000000
#define MAXBAD 8
int sIbuf[MAXBUFSIZE], rIbuf[MAXBUFSIZE];

void main(int argc, char **argv)
{
  struct sockaddr_in serverAddress;
  struct sockaddr_in addr;
  struct hostent *hp;

  int wraps, start, end, incr, t, iter, BufSize, len, eflag;
  int addrLength, sd, cnt, errcnt=0, i, j, k;
  int isServer;
  int mySocket;
  char *Sbuf, *Rbuf;

  if (argc < 2) {
    printf("To start server: %s -s \n", argv[0]);
    printf("To start client: %s server wraps start end increment\n", argv[0]);
    exit(0);
  }

  if (strcmp(argv[1], "-s")==0) { /* server */
    isServer=1;
  }
  else if (argc>=6) { /* client */
    isServer=0;
  }
  else {
    printf("To start server: %s -s \n", argv[0]);
    printf("To start client: %s server wraps start end increment\n", argv[0]);
    exit(0);
  }

  /* winsock initialization */
  printf("FD_SETSIZE=%d\n",FD_SETSIZE);
  InitWinSock();

  /* create a socket */
  fprintf(stdout,"calling socket()\n");
	fflush(stdout);
  mySocket = socket(AF_INET, SOCK_STREAM, 0);
  fprintf(stderr,"returned from socket() == %d \n",mySocket);
  if (mySocket == -1) {
	  perror("server calling sn_socket");
	  exit(-1);
  }

  /* get server and client connected */
  if (isServer) {
    /* bind the socket for server */
    serverAddress.sin_family = AF_INET;
    serverAddress.sin_addr.s_addr = INADDR_ANY;
    serverAddress.sin_port = htons(SERVER_PORT);
    fprintf(stderr,"calling bind()\n");
    if (bind(mySocket, (struct sockaddr*) &serverAddress, sizeof(struct sockaddr_in))) {
	    perror("server calling sn_bind()");
	    exit(-1);
    }

    /* listen for the client */
    listen(mySocket, 5);
  
    addrLength = sizeof(addr);
    sd = accept(mySocket, (struct sockaddr*)&addr, &addrLength);
    if (sd <= 0 ) {
	    perror("server calling sn_accept()");
	    exit(-1);
    }
  }
  else { /* client */
    /* connect to server */
    sd=mySocket;
    hp = gethostbyname(argv[1]);
  	serverAddress.sin_family = AF_INET;
    serverAddress.sin_port = htons(SERVER_PORT);
    memcpy((char *) &serverAddress.sin_addr, hp->h_addr, sizeof(serverAddress.sin_addr));
    if (connect(sd, (struct sockaddr*)&serverAddress, sizeof(struct sockaddr_in))) {
		fprintf(stderr, "client calling sn_connect returned %d :: %s\n",
			WSAGetLastError(), strerror(WSAGetLastError()));
  	  exit(-1);
    }
  }

  /* client sends the arguments to server */
  if (isServer) {
    recv(sd, &wraps, sizeof(int), 0);
    recv(sd, &start, sizeof(int), 0);
    recv(sd, &end,   sizeof(int), 0);
    recv(sd, &incr,  sizeof(int), 0);
    t = 14;
    cnt = send(sd, &t, sizeof(int), 0);
    assert(cnt == sizeof(int));
  }
  else { /* client */
    wraps = atoi(argv[2]);
    start = atoi(argv[3]);
    end   = atoi(argv[4]);
    incr  = atoi(argv[5]);
    send(sd, &wraps, sizeof(int), 0);
    send(sd, &start, sizeof(int), 0);
    send(sd, &end,   sizeof(int), 0);
    send(sd, &incr,  sizeof(int), 0);
    cnt = recv(sd, &t, sizeof(int), 0);
    if (cnt != sizeof(int) || t!=14) {
	    fprintf(stderr, "Server isn't responding\n");
	    exit(-1);
    }
  }

  fprintf(stderr,"grind with: wraps=%d start=%d end=%d incr=%d\n",
	    wraps, start, end, incr);
    
  for (i=0; i<MAXBUFSIZE; i++)
  	sIbuf[i] = rIbuf[i] = 0;
  Sbuf = (char*) sIbuf;
  Rbuf = (char*) rIbuf;
  assert(Sbuf!=NULL && Rbuf!=NULL);

  for (i=0; i<end; i++)
  	Sbuf[i]= (char)i;
  BufSize = 128*1024;


  for (len=start; len<=end; len+=incr) {
	  iter = wraps * BufSize/len;
	  fprintf(stderr,"Start: len=%d loops=%d (wraps=%d)\n",len, iter, wraps);
	  for (j=0; j<iter; j++) {
      /* client sends to server */
      if (isServer) {
	      cnt =0;
        do {
		      cnt += recv(sd, Rbuf+cnt, len-cnt, 0);
        } while (cnt!=len && cnt>0);
	      if (cnt < 0) {
		      perror("server calling recv()");
		      exit(-1);
	      } 
  	    else if (cnt == 0) {
		      fprintf(stderr, "Connection closed by client\n");
		      exit(-1);
        } else { /* data */
		      eflag=0;
		      for (k=0; k<len; k++) {
		        if (Rbuf[k] != (char)k) {
			        eflag=1;
			        fprintf(stderr,"Got:[0x%2x]!=[0x%2x] --> Bad Data at byte (%d). len=%d iter=%d\n",
				        0xff&(char)Rbuf[k], 0xff&(char)k, k, len, j);
			        if (errcnt++>MAXBAD) {
			            fprintf(stderr,"Program Stopped. Too Many Errors!\n");
			            assert(0);
			        }
            }
		        Rbuf[k] = 0;  /* clear */
          }
		      if (eflag==1)
		        fprintf(stderr,"..........\n");
        }
      }
      else { /* client */
	      cnt = send(sd, Sbuf, len, 0);  
	      assert(cnt==len);
	      if (cnt<=0) {
		      perror("client calling send()");
		      exit(-1);
	      }
      }

      /* server sends to client */
      if (!isServer) { /* client */
	      cnt =0;
        do { 
		      cnt += recv(sd, Rbuf+cnt, len-cnt, 0);
        } while (cnt!=len && cnt>0);
	      if (cnt < 0) {
		      perror("client calling recv()");
		      exit(-1);
	      } 
  	    else if (cnt == 0) {
		      fprintf(stderr, "Connection closed by server\n");
		      exit(-1);
        } else { /* data */
		      eflag=0;
		      for (k=0; k<len; k++) {
		        if (Rbuf[k] != (char)k) {
			        eflag=1;
			        fprintf(stderr,"Got:[0x%2x]!=[0x%2x] --> Bad Data at byte (%d). len=%d iter=%d\n",
				        0xff&(char)Rbuf[k], 0xff&(char)k, k, len, j);
			        if (errcnt++>MAXBAD) {
			            fprintf(stderr,"Program Stopped. Too Many Errors!\n");
			            assert(0);
			        }
            }
		        Rbuf[k] = 0;  /* clear */
          }
		      if (eflag==1)
		        fprintf(stderr,"..........\n");
        }
      }
      else { /* server */
	      cnt = send(sd, Sbuf, len, 0);  
	      assert(cnt==len);
	      if (cnt<=0) {
		      perror("client calling send()");
		      exit(-1);
	      }
      }
    }
  }

  if (errcnt)
  	fprintf(stderr,"Oh no, looks BAD, %d errors occured.\n",errcnt);
  else
	  fprintf(stderr,"LOOKS GOOD\n");

  if (isServer)
    Sleep(1000);
  close(sd);  
}
