#include <winsock2.h>
#include <windows.h>
#include <assert.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>

#define SERVER_PORT 8001

static BOOL InitWinSock(void)
{
	WORD	wVersionRequested;
	WSADATA	wsaData;
	int		err; 

	/*
	 *	WinSock requires us to call WSAStartup
	 */
	wVersionRequested = MAKEWORD( 2, 0 ); 
	err = WSAStartup( wVersionRequested, &wsaData );
	if ( err != 0 ) {
		/* Tell the user that we couldn't find a usable */
		/* WinSock DLL.                                  */    
		return FALSE;
	}

	return TRUE;
}

void main(int argc, char **argv)
{
  struct sockaddr_in serverAddress;
  struct sockaddr_in addr;
  struct hostent *hp;

  int addrLength, sd, cnt, i, j, iterations, nbytes;
  int isServer;
  int mySocket;
	LARGE_INTEGER	start, stop, freq;
  double t;
  char *sendBuf, *recvBuf;

  if (argc < 2) {
    printf("To start server: %s -s \n", argv[0]);
    printf("To start client: %s server iterations [num_bytes]+\n", argv[0]);
    exit(0);
  }

  if (strcmp(argv[1], "-s")==0) { /* server */
    isServer=1;
  }
  else if (argc>=4) { /* client */
    isServer=0;
    iterations=atoi(argv[2]);
    if (iterations<0) 
      iterations=0;
  }
  else {
    printf("To start server: %s -s \n", argv[0]);
    printf("To start client: %s server_name iterations [num_bytes]+\n", argv[0]);
    exit(0);
  }

  /* winsock initialization */
  printf("FD_SETSIZE=%d\n",FD_SETSIZE);
  InitWinSock();

  /* create a socket */
  fprintf(stdout,"calling socket()\n");
	fflush(stdout);
  mySocket = socket(AF_INET, SOCK_STREAM, 0);
  fprintf(stderr,"returned from socket() == %d \n",mySocket);
  if (mySocket == -1) {
	  perror("server calling sn_socket");
	  exit(-1);
  }

  if (isServer) {
    /* bind the socket for server */
    serverAddress.sin_family = AF_INET;
    serverAddress.sin_addr.s_addr = INADDR_ANY;
    serverAddress.sin_port = htons(SERVER_PORT);
    fprintf(stderr,"calling bind()\n");
    if (bind(mySocket, (struct sockaddr*) &serverAddress, sizeof(struct sockaddr_in))) {
	    perror("server calling sn_bind()");
	    exit(-1);
    }

    /* listen for the client */
    listen(mySocket, 5);
  
    addrLength = sizeof(addr);
    sd = accept(mySocket, (struct sockaddr*)&addr, &addrLength);
    if (sd <= 0 ) {
	    perror("server calling sn_accept()");
	    exit(-1);
    }
  }
  else { /* client */
    /* connect to server */
    sd=mySocket;
    hp = gethostbyname(argv[1]);
  	serverAddress.sin_family = AF_INET;
    serverAddress.sin_port = htons(SERVER_PORT);
    memcpy((char *) &serverAddress.sin_addr, hp->h_addr, sizeof(serverAddress.sin_addr));
    if (connect(sd, (struct sockaddr*)&serverAddress, sizeof(struct sockaddr_in))) {
		fprintf(stderr, "client calling sn_connect returned %d :: %s\n",
			WSAGetLastError(), strerror(WSAGetLastError()));
  	  exit(-1);
    }
  }

  if (isServer) {
    fprintf(stderr,"start loop\n");
    while(1) {
      /* receive iteration and size */
      cnt = recv(sd, &nbytes, sizeof(int), 0);
      if (cnt<=0) {
	        fprintf(stderr, "Connection closed by client\n");
	        exit(-1);
      }
      assert(cnt== sizeof(int));
      cnt = recv(sd, &iterations, sizeof(int), 0);
      assert(cnt== sizeof(int));
      recvBuf=malloc(nbytes);
      sendBuf=malloc(nbytes);
      assert(recvBuf!=NULL && recvBuf!=NULL);
      printf("start iter %d on %d bytes\n", iterations, nbytes);

      for(j=0; j<iterations; j++) {
        /* receive a message */
        cnt=0;
        do {
          cnt += recv(sd, recvBuf+cnt, nbytes-cnt, 0);
//        printf("recv %d \n", cnt);
        } while (cnt<nbytes && cnt>0);
        if (cnt < 0) {
	        perror("server calling sn_recv()");
	        exit(-1);
	      }
	      else if (cnt == 0) {
	        fprintf(stderr, "Connection closed by client\n");
	        exit(-1);
        }
//        printf("Recv\n");

        /* send a message */
	      cnt = send(sd, sendBuf, nbytes, 0);
	      if (cnt!=nbytes) {
	        perror("server calling sn_send()");
	        exit(-1);
	      }
//        printf("Sent\n");
      }
      free(sendBuf);
      free(recvBuf);
    }
  }
  else { /* client */
  	QueryPerformanceFrequency(&freq);

    fprintf(stderr,"start loop\n");
    for(i=3; i<argc; i++) {
      nbytes=atoi(argv[i]);
      if (nbytes<=0)
        break;
      cnt=send(sd, &nbytes, sizeof(int), 0);
      assert(cnt== sizeof(int));
      cnt=send(sd, &iterations, sizeof(int), 0);
      assert(cnt== sizeof(int));
      recvBuf=malloc(nbytes);
      sendBuf=malloc(nbytes);
      assert(recvBuf!=NULL && recvBuf!=NULL);
      printf("start iter %d on %d bytes\n", iterations, nbytes);
  		QueryPerformanceCounter(&start);
      for (j=0; j<iterations; j++) {
        /* send a message */
	      cnt = send(sd, sendBuf, nbytes, 0);	
	      if (cnt!=nbytes) {
	        perror("client calling sn_send()");
	        exit(-1);
	      }
//        printf("Sent\n");
        /* recv a message */
        cnt=0;
        do {
          cnt += recv(sd, recvBuf+cnt, nbytes-cnt, 0);
        } while (cnt<nbytes && cnt>0);
	      if (cnt < 0) {
          perror("client calling sn_recv()");
	      } else if (cnt == 0) {
          fprintf(stderr, "Connection closed by server\n");
          exit(-1);
	      }
//        printf("Recv\n");
      }
		  QueryPerformanceCounter(&stop);
      free(sendBuf);
      free(recvBuf);

	    t = (double)(stop.QuadPart - start.QuadPart)/freq.QuadPart*1000000.0;

      printf("%s niter=%d nbytes=%d OneWayLat=%7.3f usec Band=%7.3f MB/s\n",
	           argv[0],
	           iterations,
	           nbytes,
	           t/2.0/iterations,
	           2.0*nbytes*iterations/(t/1000000)/1024.0/1024.0
	       );
      }
  }
  getchar();
  close(sd);
}
